from pemap import block
from pemap import items
from pemap import value


__all__ = [
    "create_item",
    "create_block",
    "create_deep_block",
    "create_mapping",

    "items_to_tuple",
    "items_to_dict",

    "flatten_items",
    "extract_objects",
    "sort_items_by_value",

    "find_items_by_values",
    "find_item_by_values",

    "find_items_by_type",
    "find_item_by_type",

    "find_true_item",
    "find_true_items",
    "find_false_item",
    "find_false_items"
]

# Default value to use when value not provided.
default_value = value.Value.get_default_value()

def create_item(_object, value=default_value, **kwargs):
    '''Creates item object containing object and its value.

    _reference: Reference
        Instance of Reference type or any python object.
    value: Any
        Any object can sorted or support comparison operators.   
    _type: Type
        Type of reference item expectes, default: object.
    strict: Bool
        Forces `_reference` argumnet to be strictly Reference instance.
    '''
    return items.Item(_object, value, **kwargs)

def create_block(items, **kwargs):
    '''Creates block object with value from items.

    items: Iterator
        Collection of Item objects
    value: Any
        Any object can sorted or support comparison operators.   
        It needs to be compatible with items values unless 
        `update_values` is False.
    _type: Type
        Type of items this block expectes, default: object
    strict: Bool
        Prevents block from containing items containing other blocks.
    value_mode: Str
        Mode for calculating value for block and items, default:
        'median'.
    update_values: Bool
        Enables and disables updating of block and items values.
    
    If items contains block object consider using `create_deep_block()`
    as it will extract the items of that block. Continue using this
    function if thats what you want.'''
    return block.Block(items, **kwargs)

def create_deep_block(items, **kwargs):
    '''Creates deep block object from items. 

    items: Iterator
        Collection of Item objects
    value: Any
        Any object can sorted or support comparison operators.   
        It needs to be compatible with items values unless 
        `update_values` is False.
    _type: Type
        Type of items this block expectes, default: object
    strict: Bool
        Prevents block from containing items containing other blocks.
    value_mode: Str
        Mode for calculating value for block and items, default:
        'median'.
    update_values: Bool
        Enables and disables updating of block and items values.
    
    Deep block is neccessay when items can contain block object
    which may contain other items. This function results in block object
    that contain other items extracted. Any nested block objects within
    items get removed while retaining non block items.
    
    This is different from `create_block()` as it keeps items containing
    block object. That makes it harder to access deep or low-level items
    within the nested block objects.

    Note that extracted deep/low-level items are only copies of original.
    Block object first copies items before performing anything on them.
    Getting items throuh `.get_items()` will only return the copies.
    '''
    return block.DeepBlock(items, **kwargs)


def create_mapping(items, flatten=False, **kwargs):
    '''Creates corresponding block object based on 'flatten' argument.

    items: Iterator
        Collection of Item objects
    flatten: Bool
        Enables and disables use `create_block()` or `create_deep_block()`.
    value: Any
        Any object can sorted or support comparison operators.   
        It needs to be compatible with items values unless 
        `update_values` is False.
    _type: Type
        Type of items this block expectes, default: object
    strict: Bool
        Prevents block from containing items containing other blocks.
    value_mode: Str
        Mode for calculating value for block and items, default:
        'median'.
    update_values: Bool
        Enables and disables updating of block and items values.

    When 'flatten' is True, `create_deep_block()` will be used to create 
    block object else `create_block()`. Set 'flatten' argument to 
    True to remove any nested block objects and replace them with their 
    items.
    '''
    if flatten:
        return create_deep_block(items, **kwargs)
    else:
        return create_block(items, **kwargs)


######################################################################
# Functions defined after here internally creates block object.
# It may be better to manually create block object for performance.
# These functions are meant to give functional programming flavour.
# Manually creating block object could result in few more advantages.
######################################################################

def items_to_tuple(items, flatten=False):
    '''Convert items into map like tuple'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.to_tuple()

def items_to_dict(items, flatten=False):
    '''Convert items into multi dict'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.to_dict()



def flatten_items(items):
    '''Flattens items by exposing items within nested block objects.
    
    This function removes any block object within items while retaining
    items. The number of resulting items may be larger than items 
    provided on argumnets. This is because this function exposes items
    that were nested deep within block objects.
    
    This function does not affect nested items but items containing
    block. Nested items is something out of scope of this library but block
    and items can be nested.

    Block can contain items containing other blocks and items can contain
    block objects. But items containg other items is something that wasnt
    planned for this library.
    '''
    block_object = create_deep_block(items)
    return block_object.get_items()

def sort_items_by_value(items):
    '''Sorts items based on their values'''
    block_object = create_block(items, strict=False)
    return block_object.get_sorted_items()

def extract_objects(items, flatten=False):
    '''Extracts objects within items'''
    #return [_item.get_object() for _item in items]
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_objects()



def find_items_by_values(items, values, flatten=False):
    '''Finds items with values matching any of values'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_items_by_values(values)

def find_item_by_values(items, values, flatten=False):
    '''Finds item with value matching any of values'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_item_by_values(values)



def find_items_by_type(items, _type, flatten=False):
    '''Finds items with type matching provided type'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_items_by_type(_type)

def find_item_by_type(items, _type, flatten=False):
    '''Finds item with type matching provided type'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_item_by_type(_type)

def find_true_items(items, flatten=False):
    '''Gets items that evaluates to true.'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_true_items()

def find_true_item(items, flatten=False):
    '''Gets first item evaluating to true.'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_true_item()

def find_false_items(items, flatten=False):
    '''Gets items that evaluates to false.'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_false_items()

def find_false_item(items, flatten=False):
    '''Gets first item evaluating to false.'''
    block_object = create_mapping(items, flatten=flatten, strict=False)
    return block_object.get_false_item()


from pemap.reference import Reference
from pemap.value import Value

from pemap.items import BaseItem
from pemap.items import Item

from pemap.block import BaseBlock
from pemap.block import Block
from pemap.block import DeepBlock

from pemap.highlevel import *


__name__ = "pemap"
___version__ = "0.2.0"



import collections
import inspect


class Reference():
    # Wraps object and provide methods for operating on it.
    def __init__(self, _object) -> None:
        # _object: Any python object
        self._object = _object

    def set_object(self, _object):
        self._object = _object

    def get_object(self):
        return self._object

    def get_type(self):
        # returns type of underlying object
        return self._object.__class__

    @classmethod
    def to_reference(cls, _object):
        # Creates refence object from provided object.
        if isinstance(_object, Reference):
            _item = _object
        else:
            _item = cls(_object)
        return _item

    def is_callable(self):
        # Checks if underlying object is callable
        return callable(self._object)

    def is_method_func(self):
        # Checks if object if value is function or method.
        if self.is_callable():
            if inspect.isfunction(self._object):
                return True
            else:
                return inspect.ismethod(self._object)
        return False

    def is_iterable(self):
        # Checks if underlying object is iterable
        return isinstance(self._object, collections.Iterable)

    def is_iterator(self):
        # Checks if underlying object is iterator
        return isinstance(self._object, collections.Iterator)

    def is_string(self):
        # Checks if underlying object is string
        return isinstance(self._object, str)

    def is_bytes(self):
        # Checks if underlying object is bytes
        return isinstance(self._object, str)

    def is_string_bytes(self):
        # Checks if underlying object is string or bytes
        return self.is_string() or self.is_bytes()





def rename_attribute(object_, old_attr_name, new_attr_name):
    # Renames attribute by replacing it with new name.
    setattr(object_, new_attr_name, getattr(object_, old_attr_name))
    delattr(object_, old_attr_name)

def copy_attribute(object_, old_attr_name, new_attr_name):
    # Creates copy of attribute with new name.
    setattr(object_, new_attr_name, getattr(object_, old_attr_name))




from pemap import reference

    
class Value(reference.Reference):
    '''Defines value to be used with item'''
    # Used for error messages and others.
    _name = "value"
    # Default value when value is not provided.
    # This is taken as if value is not provided(careful).
    # Think of it as None with arguments of function.
    _default_value = ...

    # _value_attr_names = (object attr, object method)
    # Used for getting value for object.
    _default_value_attr_names = ("value", "get_value") 
    _value_attr_names = _default_value_attr_names

    def __init__(self, value) -> None:
        super().__init__(value)

    @classmethod
    def get_name(cls):
        return cls._name

    @classmethod
    def get_default_value(cls):
        return cls._default_value

    @classmethod
    def get_value_attr_names_names(cls):
        return cls._value_attr_names

    @classmethod
    def get_value_attr_name(cls):
        return cls._value_attr_names[0]

    @classmethod
    def get_value_method_name(cls):
        return cls._value_attr_names[1]

    def get_value(self, *args, **kwargs):
        # Gets value behind this object.
        # Object of method or function will result in return value
        # of that callable.
        if self.is_method_func():
            return self._object(*args, **kwargs)
        elif isinstance(self._object, Value):
            return self._object.get_value()
        else:
            return self._object

    def set_value(self, value):
        self._object = value





from pemap import items as items_
from pemap import value as value_

from collections import defaultdict


class BaseBlock():
    '''Wraps collection of Item objects and associate them with value.'''
    def __init__(self, items, _type=object) -> None:
        self._items = items
        self._type = _type

    def _to_items(self, _items_like):
        # Returns item objects from iterator of objects.
        # Item objects will be returned unchanged.
        # Non item objects will  result in item objects.
        return  [items_.Item.to_item(_item) for _item in _items_like]

    def copy_items(self):
        # Copies current items of block
        return [self._item.copy() for _item in self._items]

    def get_items(self):
        # Returns items stored in block object
        return self._items
    
    def get_objects(self, value_sort=False):
        '''Gets items underlying objects'''
        return self.extract_objects_from_items(self._items)

    @classmethod
    def extract_objects_from_items(cls, items):
        '''Gets underlying object from item object'''
        return [_item.get_object() for _item in items]

    @classmethod
    def sort_items_by_value(cls, items):
        '''Returns items sorted by their values'''
        return sorted(items, key=lambda _item: _item.get_value())

    def filter_items(self, key=None, limit=None):
        '''Filters item objects filtered by key function'''
        if limit == None: 
            items = self._items
        else:
            items = self._items[:limit]
        return list(filter(key, items))



    def get_values(self):
        '''Gets values of block item objects'''
        return [_item.get_value() for _item in self._items]

    def get_items_by_value(self, value):
        '''Gets item objects matching value'''
        def func(_item):
            return _item.get_value() == value
        return self.filter_items(func)

    def get_item_by_value(self, value):
        '''Gets first item matching value'''
        items = self.get_items_by_value(value)
        if items: return items[0]

    def get_items_by_values(self, values):
        '''Gets item objects matching any of values'''
        def func(_item):
            return _item.get_value() in values
        return self.filter_items(func)

    def get_item_by_values(self, values):
        '''Gets first item matching any of values'''
        items = self.get_items_by_values(values)
        if items: return items[0]

    def get_items_by_type(self, _type):
        '''Gets item objects of provided type'''
        # Type is defined as type of object underlying item.
        def func(_item):
            return isinstance(_item.get_value(), _type)
        return self.filter_items(func)   

    def get_item_by_type(self, _type):
        '''Gets first item of provided type'''
        items = self.get_items_by_type(_type)
        if items: return items[0] 


    def get_true_items(self):
        # Gets items that evaluates to true.
        return self.filter_items(lambda item: item.get_value())

    def get_true_item(self):
        # Gets first item evaluating to true.
        items = self.filter_items(lambda item: item.get_value(), 1)
        if items: return items[0]

    def get_false_items(self):
        # Gets items that evaluates to false.
        return self.filter_items(lambda item: not item.get_value())

    def get_false_item(self):
        # Gets first item evaluating to false.
        items = self.filter_items(lambda item: not item.get_value(), 1)
        if items: return items[0]



    def to_tuple(self):
        '''Returns tuple form of block with values and objects'''
        # Value will be used as tuple key and object as value.
        # object is the object under reference object of items
        results = []
        for _item in self._items:
            results.append((_item.get_value(), _item.get_object()))
        return tuple(results)

    def to_dict(self):
        '''Returns dict form of block with values and objects'''
        # This will fail if value not hashable.
        return dict(self.to_tuple())

    def _to_multi_dict(self):
        # Returns multi dict from items values and underlying objects
        # Key is value and values are underlying objects.
        result_dict = defaultdict(set)
        for _value, _object in self.to_tuple():
            result_dict[_value].add(_object)
        return result_dict

    def to_multi_dict(self):
        '''Returns multi dict from items values and underlying objects.
        Key is value and values are underlying objects.'''
        return self._to_multi_dict()

    def __iter__(self):
        return iter(self.get_sorted_items())

    def __len__(self):
        return len(self._items)


class Block(BaseBlock):
    '''Wraps collection of Item objects and associate them with value.
    
    Instances of this class first copies each item before performing
    any operation on them. Value for item can influence value
    for block instance if value for block is not provided.
    
    When `strict` is True, block instance will not allow item containing
    another block. This is by default set to True to avoid confusion
    but can be set to True to allow nested block instances.'''
    # Type of item type items of block are expected.
    _item_type = items_.Item

    def __init__(self, items, _type=object, strict=True):
        '''
        items: Iterator
            Collection of Item objects
        value: Any
            Any object can sorted or support comparison operators.   
            It needs to be compatible with items values unless 
            `update_values` is False.
        _type: Type
            Type of items this block expectes, default: object
        strict: Bool
            Prevents block from containing items containing other blocks.
        value_mode: Str
            Mode for calculating value for block and items, default:
            'median'.
        update_values: Bool
            Enables and disables updating of block and items values.
        '''
        super().__init__(items, _type)
        self._items = items
        self._type = _type
        self._strict = strict
        # Setup items after value have been set from existing items.
        # The method may modify values for items.
        # Value is passed as argument since value argument may
        # have been modified.
        # This are results of calling methods within initialiser.
        self._setup_items(items)
        # Calling method within initializer is hell.
        # The instance is not yet fully created.
        # Warning has been included on the methods called by __init__().

    def _setup_items(self, items):
        # Setup items to ensure they are in correct type.
        # Item objects will be created when neccessary.
        # This could make find bugs hard but it simplifies things.
        # This method is not meant to be overiden(take care)
        new_items = []
        for _item in items:
            new_item = items_.Item.to_item(_item)
            # Gets object underlying item.
            _object = new_item.get_object()
            # Check if strict is respected(Block objects not allowed).
            # Exception is raised if not respected.
            if self._strict and isinstance(_object, Block):
                err_msg = "Nested Block objects not allowed when " +\
                    "'strict' is enabled"
                raise TypeError(err_msg)
            # Check if type for object is correct.
            # Exception is if type of object does not match expected one.
            if not isinstance(_object, self._type):
                err_msg = "Item should have reference of type '{}' not '{}'"
                type_name = _object.__class__.__name__
                err_msg = err_msg.format(self._type.__name__, type_name)
                raise TypeError(err_msg)
            # Appends new item to items list
            new_items.append(new_item)
        self._items = new_items
        #self._items_dict = dict(self._to_multi_dict())



class DeepBlock(Block):
    ''' Varient of Block that allows extracting of deep/low-level items.

    This block class differs from its parent in that it eliminates
    nested block instances by extracting their items and then removing
    any item that contain block instances.
    
    At end of day, items with block instances are removed while retaining
    their items. This class gurantees that its items wont have any block
    object. That makes it easy to access items that were deep into nested
    block objects.
    
    `strict` is now set to False as this block type is based on nested 
    block objects. One should now not fear to nest block objects together
    within item objects that will be used on another block object.
    
    values for block and items will be updated accordinly as similar
    to its parent class.'''
    def __init__(self, items, _type=object, strict=False):
        super().__init__(items, _type, strict)
    
    @classmethod
    def _extract_deep_items(cls, _block):
        # Extracts low level(deep) items from block object.
        # This include item objects not containing block object.
        # This method is called by _setup_items().
        # Take care when extensing it on sub classes.
        deep_items = []
        # Gets item objects of block(not sorted by value)
        items = _block.get_items()
        for _item in items: 
            # Gets value for item.
            _object = _item.get_value() 
            # Checks if object of reference is block object.
            # Uses recursion if the object is block object.
            # Recursion continues until non block item is found.
            if isinstance(_object, Block):
                # Extract non block items from the block object.
                block_deep_items = cls._extract_deep_items(_object)
                deep_items.extend(block_deep_items)
            else:
                # This item does not contain block object
                deep_items.append(_item)
        return deep_items

    def _setup_items(self, items,):
        # Setup deep items overiding existing item objects.
        # Ensures all items are really item objects.
        # _extract_deep_items() expectes item objects.
        self._items = self._to_items(self._items)
        _items = self._extract_deep_items(self)
        # Now asks super class to setup items as usual.
        # Items values will be updated as expected.
        super()._setup_items(_items)


if __name__ == "__main__":
    from pemap import reference

    item_object = items_.Item(reference.Reference("Ruth"), "b")
    item_object2 = items_.Item("Marry", "a")
    item_object3 = items_.Item("John", "c")

    items = [item_object, item_object2, item_object3]

    block_object = Block(items)
    block_item = items_.Item(block_object, "we")
    deep_object = DeepBlock([block_item])

    print("block value: ", block_object.get_values())
    print("deep_block value: ", deep_object.get_values())


from pemap import reference as reference_
from pemap import value as value_


class BaseItem():
    '''Associates reference object with any python object/value.
    
    Instances of this class store a reference containing object 
    and associate it with value. If non reference object
    is passed as reference, reference object will be created form it.
    
    Value defines value of reference over other references. Value can
    be anything or any python object.
    
    Remember that this class internally creates reference object if 
    reference argument is not Reference object.'''
    # Value class/type to use.
    _value_type = value_.Value

    def __init__(self, reference, value=_value_type.get_default_value(), 
    _type=object,  strict=False):
        '''
        reference: Reference
            Instance of Reference type or any python object.
        Value: Any
            Any object to be associated with item.  
        _type: Type
            Type of reference item expectes, default: object.
        strict: Bool
            Forces `reference` argumnet to be strictly Reference instance.
        '''
        self._type = _type
        self._strict = strict
        self._type = _type
        # Calling method within initializer can cause problems.
        # This should be avoided as possible to avoid hard issues.
        # __init__() should be for initialising data not method calls.
        self._setup_reference(reference)
        # Shouldnt super().__init__() be first call of __init__()?
        # Thats because super().__init__() calls another method which
        # is overided by this class(begining of problems).
        self._setup_value(value)

    @classmethod
    def get_value_type(cls):
        return cls._value_type

    @classmethod
    def get_value_name(cls):
        return cls._value_type.get_name()

    def _setup_value(self, value):
        # Setup value for item.
        # This method is not meant to be overiden(take care)
        if value == self._value_type.get_default_value():
            # Sets up variables to be used to hget value
            _object = self.get_object()
            value_attr_name = self._value_type.get_value_attr_name()
            value_method_name = self._value_type.get_value_method_name()
            # Attempt to get value from object.
            if hasattr(_object, value_attr_name):
                value = getattr(_object, value_attr_name)
            elif hasattr(_object, value_method_name):
                value = getattr(_object, value_method_name)()
            else:
                err_msg = "Cannot get {0} from object of type " +\
                    "'{1}', please provide {0} or define " +\
                    "'{2}()' or '{3}' attributes."
                # Set string format variables
                type_name = _object.__class__.__name__
                value_name = self._value_type.get_name()
                # Format string with those variables.
                err_msg = err_msg.format(value_name, type_name, value_method_name, 
                value_attr_name)
                raise AttributeError(err_msg)
        # Value for item cann
        if self._value_type.get_default_value() == value:
            type_name = _object.__class__.__name__
            value_name = self._value_type.get_name().capitalize()
            err_msg = "{} for item cannot be '{}'"
            raise ValueError(err_msg.format(value_name, type_name))
        # Now set value attribute using Value type.
        self._value = self._value_type(value)

    def _setup_reference(self, reference):
        # Creates reference object when neccessay
        # This method is not meant to be overiden(take care)
        if self._strict and not isinstance(reference, reference_.Reference):
                err_msg = "Reference needs to be instance of '{}' not " +\
                    "'{}' when 'strict' is enabled"
                err_msg = err_msg.format(
                    reference.Reference.__name__,
                    reference.__class__.__name__
                )
                raise TypeError(err_msg)
        else:
            self._reference = reference_.Reference.to_reference(reference)
            if not isinstance(reference, self._type):
                err_msg = "object should be type {}, not {}"
                raise TypeError(err_msg.format(reference, self._type))

    @classmethod
    def to_item(cls, _object):
        # Creates item object from if not already item object
        if isinstance(_object, Item):
            _item = _object
        else:
            _item = cls(_object)
        return _item

    def set_value(self, value):
        # Sets value/object behind item.
        return self._value_type(value)
    
    def get_value(self, *args, **kwargs):
        # Gets value behind this item.
        return self._value.get_value(*args, **kwargs)

    def get_reference(self):
        # Gets underling reference object
        return self._reference

    def get_type(self):
        # Gets type of underlying reference
        return self._reference.get_type()

    def get_object(self):
        # Gets object of underlying reference
        return self._reference.get_object()

    def copy(self):
        # Creates a copy of item
        return self.__class__(self._reference, self._value, self._type, 
        self._strict)


class Item(BaseItem):    
    def __init__(self, reference, value=..., *args, **kwargs):
        super().__init__(reference, value, *args, **kwargs) 


if __name__ == "__main__":
    item = Item(10, lambda :34)
    item2 = Item(item)
    item3 = Item(item2)
    print(item.get_object())
    print(item3.get_value())